/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "Label.h"
#include "LabelSettings.h"
#include "xFont.h"
#include "SystemInfo.h"

#include "VerInfo.h"

#ifndef xLabelLight
//----------------------------------------------------------------
	#include "FileInformation.h"
	#include "NotifyDirCheck.h"
//----------------------------------------------------------------
#endif

#include <fstream>

//DropShadow under XP
#ifndef CS_DROPSHADOW
#define CS_DROPSHADOW 0x00020000
#endif

//PlaySound
#define   NOCOMPMAN     //  - dont include COMPMAN
#define   NODRAWDIB   //    - dont include DRAWDIB
#define   NOVIDEO     //    - dont include video capture interface
#define   NOAVIFMT    //    - dont include AVI file format structs
#define   NOMMREG     //    - dont include MMREG
#define   NOAVIFILE   //    - dont include AVIFile interface
//#define   NOMCIWND    //    - dont include AVIWnd class.
#define   NOAVICAP    //    - dont include AVICap class.
#define   NOMSACM     //    - dont include ACM stuff.
#include "vfw.h"

// Bang commands
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs);
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangDock(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs);
void BangMove(HWND hwndCaller, LPCSTR pszArgs);
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs);
void BangNext(HWND hwndCaller, LPCSTR pszArgs);
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs);
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs);
void BangReposition(HWND hwndCaller, LPCSTR pszArgs);
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs);
void BangResize(HWND hwndCaller, LPCSTR pszArgs);
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs);
void BangScroll(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs);
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs);
void BangSetFontColor(HWND hwndCaller, LPCSTR pszArgs);
void BangSetFontShadowColor(HWND hwndCaller, LPCSTR pszArgs);
void BangSetText(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);
void BangGhosted(HWND hwndCaller, LPCSTR pszArgs);
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs);

void BangInfoExport(HWND hwndCaller, LPCSTR pszArgs);

void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs);
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs);

void BangPlaySound(HWND hwndCaller, LPCSTR pszArgs);

#ifndef xLabelLight
//----------------------------------------------------------------
	//Shortcut	
	bool useshortcut2 = GetxRCBool("xlabel", "UseShortcut2");

	void BangSCOnBottom(HWND hwndCaller, LPCSTR pszArgs);
	void BangSCToggleOnTop(HWND hwndCaller, LPCSTR pszArgs);

	//xTextedit
	#include "regexp.h"
	#include "regmagic.h"

	bool usetextedit = GetxRCBool("xlabel", "UsexTextEdit");
	bool userConfirmation = GetxRCBool("xTextEditConfirmation", "");
	bool userDebug = GetxRCBool("xTextEditDebug", "");

	void BangTextAppend(HWND hwndCaller, LPCSTR pszArgs);
	void BangTextDelete(HWND hwndCaller, LPCSTR pszArgs);
	void BangTextReplace(HWND hwndCaller, LPCSTR pszArgs);
	void BangTextReplaceAll(HWND hwndCaller, LPCSTR pszArgs);
	void BangTextSaveEvar(HWND hwndCaller, LPCSTR pszArgs);

	void getToken(char line[MAX_LINE_LENGTH], char temp[MAX_LINE_LENGTH], bool escape_codes = true);

	//xDesktop
	bool usexdesktop = GetxRCBool("xlabel", "UsexDesktop");
	bool useautodesktop = GetxRCBool("xdesktop", "AutoCreate");
	string desktopSavePath = GetxRCString("xdesktop", "ConfigFile", "");

	void BangDesktopCreate(HWND hwndCaller, LPCSTR pszArgs);
	void BangDesktopDestroy(HWND hwndCaller, LPCSTR pszArgs);
	void BangDesktopHide(HWND hwndCaller, LPCSTR pszArgs);
	void BangDesktopShow(HWND hwndCaller, LPCSTR pszArgs);
	void BangDesktopToggle(HWND hwndCaller, LPCSTR pszArgs);

	bool customDelete = false;

	StringList getActualDesktopItemPaths(bool defaultDesktop, string searchPath);
	void DesktopUpdate(int updateMode, string updatedFile, StringList ActualDesktopItemPaths);

	UINT DirCallback( CFileInformation fiObject, EFileAction faAction, LPVOID lpData );

	void WatchStart(LPCSTR path);
	void WatchStop();

	class CNotifyDirCheck;

	typedef list<CNotifyDirCheck *> DirCheckList;
	typedef list<CNotifyDirCheck *>::iterator DirCheckListIterator;

	DirCheckList dircheckList;

	CStringList pathList;
//----------------------------------------------------------------
#endif

HWND m_hWndSoundFile = NULL;

string Bang = "";

HINSTANCE hInstance;
HWND messageHandler;

LabelList labelList;
LabelList desktoplabelList;
LabelList shortcutlabelList;

#ifndef xLabelLight
//----------------------------------------------------------------
	StringList labelDesktopItems;
	StringList labelShortcutItems;
//----------------------------------------------------------------
#endif

int lsMessages[] = {
	LM_GETREVID,
	LM_REFRESH,
	0
};

Label *lookupLabel( const string &name, LabelList searchList = labelList );

#define LM_UPDATEBG (WM_USER + 1)

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}
		
		case LM_REFRESH:
		{
			StringList labelNames = GetxRCNameList("*label");

			// refresh the "AllLabels" configuration
			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if(!(*lit)->getBox())
				{
					// destroy all labels that no longer exist and that are not in a box
					for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
					{
						if(stricmp((*sit).c_str(), (*lit)->getName().c_str()) == 0)
							break;
					}
					if (sit == labelNames.end())
					{
						labelList.remove(*lit);
						delete *lit;
						continue;
					}
				}

				// we can reconfigure all other labels, even if they are "boxed"
				(*lit)->reconfigure(true);
				(*lit)->repaint(true);
			}

			// create the rest
			for(StringListIterator sit = labelNames.begin(); sit != labelNames.end(); sit++)
			{
				Label *label = lookupLabel(*sit);
				
				if (!label) 
				{
					label = new Label(*sit);
					labelList.insert(labelList.end(), label);
					if ( IsOS(OS_XP) && GetxRCBool((*sit).c_str(), "DropShadow") )
						label->load(hInstance, 0, true);
					else 
						label->load(hInstance);

					//ModuleHook 
					if (!label->moduleHook.empty())
					{
						for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
						{
							string hWndbuffer = IntToString( (int) label->gethWnd() );
							string config = (*sit).c_str();

							if ( !label->configSavePath.empty() )
							{
								config.append( " \"" );
								config.append( label->configSavePath.c_str() );
								config.append( "\" " );
								config.append( hWndbuffer );
							}
							else
							{
								config.append( " \"$themedir$theme.rc\" " );
								config.append( hWndbuffer );
							}

							LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
						}
					}

					if (!label->isStartHidden())
						label->show();
					else
						label->hide();
					
					//OnResize one time after all are loaded and shown
					//if( !label->resizeCommand.empty() )
					//	xLSExecute(label->gethWnd(), label->resizeCommand.c_str(), SW_SHOWNORMAL);
				}
			}

			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			LabelListIterator lit;
			for(lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				Label *label = *lit;

				if(label->getBox() == 0 && label->getTransparencyMode() == 2)
					label->repaint(true);
			}

			#ifndef xLabelLight
			//----------------------------------------------------------------
			for(lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				Label *label = *lit;

				if(label->getBox() == 0)
					label->repaint(true);
			}
			//----------------------------------------------------------------
			#endif

			return 0;
		}

		case WM_DISPLAYCHANGE:

		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return 0;
		}

		case WM_TIMER:
		{
			if (m_hWndSoundFile != NULL)
			{
				if (MCIWndGetEnd(m_hWndSoundFile) == MCIWndGetPosition(m_hWndSoundFile))
				{
					MCIWndStop(m_hWndSoundFile);
					MCIWndDestroy(m_hWndSoundFile);
					m_hWndSoundFile = NULL;

					KillTimer(messageHandler, NULL);
				}
			}
			else
				KillTimer(messageHandler, NULL);
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes);

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);

	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS ;
	wc.lpfnWndProc = Label::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Label *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabel";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	if(IsOS(OS_XP))
	{
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = CS_GLOBALCLASS | CS_DBLCLKS | CS_DROPSHADOW;
		wc.lpfnWndProc = Label::windowProcedure;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = sizeof(Label *);
		wc.hInstance = hInstance;
		wc.hbrBackground = 0;
		wc.hCursor = 0;
		wc.hIcon = 0;
		wc.lpszMenuName = 0;
		wc.lpszClassName = "xLabelXP";
		wc.hIconSm = 0;

		RegisterClassEx(&wc);
	}

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xLabelMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xLabelMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	defaultSettings = new LabelSettings();
	systemInfo = new SystemInfo();

	AddBangCommand("!LabelAlwaysOnTop", BangAlwaysOnTop);
	AddBangCommand("!LabelClipboardCopy", BangClipboardCopy);
	AddBangCommand("!LabelClipboardPaste", BangClipboardPaste);
	AddBangCommand("!LabelCreate", BangCreate);
	AddBangCommand("!LabelDestroy", BangDestroy);
	AddBangCommand("!LabelDock", BangDock);
	AddBangCommand("!LabelHide", BangHide);
	AddBangCommand("!LabelLSBoxHook", BangLSBoxHook);
	AddBangCommand("!LabelMove", BangMove);
	AddBangCommand("!LabelMoveBy", BangMoveBy);
	AddBangCommand("!LabelNext", BangNext);
	AddBangCommand("!LabelPrevious", BangPrevious);
	AddBangCommand("!LabelRefresh", BangRefresh);
	AddBangCommand("!LabelReposition", BangReposition);
	AddBangCommand("!LabelRepositionBy", BangRepositionBy);
	AddBangCommand("!LabelResize", BangResize);
	AddBangCommand("!LabelResizeBy", BangResizeBy);
	AddBangCommand("!LabelScroll", BangScroll);
	if(IsOS(OS_2KXP))
		AddBangCommand("!LabelSetAlpha", BangSetAlpha);
	AddBangCommand("!LabelSetAnimation", BangSetAnimation);
	AddBangCommand("!LabelSetFontColor", BangSetFontColor);
	AddBangCommand("!LabelSetFontShadowColor", BangSetFontShadowColor);
	AddBangCommand("!LabelSetText", BangSetText);
	AddBangCommand("!LabelShow", BangShow);
	AddBangCommand("!LabelShowHide", BangShowHide);
	AddBangCommand("!LabelToggle", BangToggle);
	AddBangCommand("!LabelGhosted", BangGhosted);
	AddBangCommand("!LabelUpdate", BangUpdate);

	AddBangCommand("!LabelInfoExport", BangInfoExport);

	AddBangCommand("!ParseEvars", BangParseEvars);
	AddBangCommand("!SetEvar", BangSetEvars);

	AddBangCommand("!PlaySound", BangPlaySound);

	#ifndef xLabelLight
	//----------------------------------------------------------------
	//xTextEdit
	if (usetextedit)
	{
		AddBangCommand("!xTextAppend", BangTextAppend);
		AddBangCommand("!xTextDelete", BangTextDelete);
		AddBangCommand("!xTextReplace", BangTextReplace);
		AddBangCommand("!xTextReplaceAll", BangTextReplaceAll);
		AddBangCommand("!xTextSaveEvar", BangTextSaveEvar);
	}

	//Shortcut2
	if (useshortcut2)
	{
		AddBangCommand("!ShortcutGroupHide", BangHide);
		AddBangCommand("!ShortcutGroupShow", BangShow);
		AddBangCommand("!ShortcutGroupToggle", BangToggle);
		AddBangCommand("!ShortcutGroupOnTop", BangAlwaysOnTop);
		AddBangCommand("!ShortcutGroupOnBottom", BangSCOnBottom);
		AddBangCommand("!ShortcutGroupOnTopToggle", BangSCToggleOnTop);
	}

	//xDesktop
	if (usexdesktop)
	{
		AddBangCommand("!xDesktopCreate", BangDesktopCreate);
		AddBangCommand("!xDesktopDestroy", BangDesktopDestroy);
		AddBangCommand("!xDesktopHide", BangDesktopHide);
		AddBangCommand("!xDesktopShow", BangDesktopShow);
		AddBangCommand("!xDesktopToggle", BangDesktopToggle);
	}
	//----------------------------------------------------------------
	#endif

	//Check LoadModule Order
	if ( FindWindow("DesktopBackgroundClass", NULL) == NULL && !GetxRCBool("xLabel", "NoDesktopWarning"))
		MessageBox(NULL, "Just To Get Rid Of One Ever Returning \"Bug Report\"!\n\nPlease Load A DesktopModule (Desktop2 or jDesk) BEFORE xLabel(Light)!\nOtherwise You Will Get Annoying Z-Order Problems!\n\nBtw.: This Message Will Pop Up On Every Recycle Till You Fix It :P", "Important User Information", MB_SETFOREGROUND | MB_ICONEXCLAMATION );

	StringList labelNames = GetxRCNameList("*label");

	StringListIterator sit;

	for(sit = labelNames.begin(); sit != labelNames.end(); sit++)
	{
		Label *label = new Label(*sit);
	
		labelList.insert(labelList.end(), label);
		
		if ( IsOS(OS_XP) && GetxRCBool((*sit).c_str(), "DropShadow") )
			label->load(hInstance, 0, true);
		else 
			label->load(hInstance);
	}

	#ifndef xLabelLight
	//----------------------------------------------------------------
	//xDesktop Items
	if (usexdesktop)
	{
		labelDesktopItems = GetxRCLineList("*xdesktopitem");

		if (!labelDesktopItems.empty())
		{	
			for(sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
			{
				char name[32];
				char *buffers[] = {name};

				if(LCTokenize((*sit).c_str(), buffers, 1, 0) >= 1)
				{
					Label *label = new Label(name);

					desktoplabelList.insert(desktoplabelList.end(), label);

					label->desktopload(hInstance, (*sit));
				}
			}
		}
	}
	 
	//Shortcut Items
	if (useshortcut2)
	{
		labelShortcutItems = GetxRCLineList("*shortcut");

		if (!labelShortcutItems.empty())
		{
			int lastx = 0;
			int lasty = 0;
			for(sit = labelShortcutItems.begin(); sit != labelShortcutItems.end(); sit++)
			{
				char name[64], ix[16], iy[16];
				char *buffers[] = {name, ix, iy};

				if(LCTokenize((*sit).c_str(), buffers, 3, 0) >= 3)
				{
					Label *label = new Label(name);

					shortcutlabelList.insert(shortcutlabelList.end(), label);

					label->shortcutload(hInstance, (*sit), lastx, lasty);

					lastx = xParseCoord(ix, 0, GetSystemMetrics(SM_CXSCREEN), lastx);
					lasty = xParseCoord(iy, 0, GetSystemMetrics(SM_CYSCREEN), lasty);		
				}
			}
		}
	}
	//----------------------------------------------------------------
	#endif

	//Initial Show
	LabelListIterator lit;

	for(lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		//ModuleHook
		if (!(*lit)->moduleHook.empty())
		{
			for(StringListIterator sit = (*lit)->moduleHook.begin(); sit != (*lit)->moduleHook.end(); sit++)
			{
				string hWndbuffer = IntToString( (int) (*lit)->gethWnd() );
				string config = (*sit).c_str();

				if ( !(*lit)->configSavePath.empty() )
				{
					config.append( " \"" );
					config.append( (*lit)->configSavePath.c_str() );
					config.append( "\" " );
					config.append( hWndbuffer );
				}
				else
				{
					config.append( " \"$themedir$theme.rc\" " );
					config.append( hWndbuffer );
				}

				LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
			}
		}

		if (!(*lit)->isStartHidden())
			(*lit)->show();
		else
			(*lit)->hide();
	}

	for(lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		for(StringListIterator sit = (*lit)->dockedToList.begin(); sit != (*lit)->dockedToList.end(); sit++)
		{
			for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
			{
				if( stricmp( (*sit).c_str(), (*plit)->getName().c_str()) == NULL )
				{
					(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), (*lit));
					break;
				}
			}
		}
	}

	for(lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		(*lit)->dockedToList.clear();
	}

	#ifndef xLabelLight
	//----------------------------------------------------------------
		if (usexdesktop)
		{
			if (useautodesktop)
			{
				if (labelDesktopItems.empty())
				{
					Bang = "!xdesktopcreate replace \".desktop\"";
					LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
				}

				char path[MAX_PATH_LENGTH];			
				VarExpansion( path, "$desktopdir$" );
				WatchStart( path );
				VarExpansion( path, "$commondesktopdir$" );
				WatchStart( path );
			}
		}
	//----------------------------------------------------------------
	#endif

	return 0;
}

extern HDC hdcDesktop;
extern HBITMAP hbmDesktop;

void quitModule(HINSTANCE hInstance)
{
	if (m_hWndSoundFile != NULL)
	{
		MCIWndStop(m_hWndSoundFile);
		MCIWndDestroy(m_hWndSoundFile);
		m_hWndSoundFile = NULL;
	}

	RemoveBangCommand("!LabelAlwaysOnTop");
	RemoveBangCommand("!LabelClipboardCopy");
	RemoveBangCommand("!LabelClipboardPaste");
	RemoveBangCommand("!LabelCreate");
	RemoveBangCommand("!LabelDestroy");
	RemoveBangCommand("!LabelDock");
	RemoveBangCommand("!LabelHide");
	RemoveBangCommand("!LabelLSBoxHook");
	RemoveBangCommand("!LabelMove");
	RemoveBangCommand("!LabelMoveBy");
	RemoveBangCommand("!LabelNext");
	RemoveBangCommand("!LabelPrevious");
	RemoveBangCommand("!LabelRefresh");
	RemoveBangCommand("!LabelReposition");
	RemoveBangCommand("!LabelRepositionBy");
	RemoveBangCommand("!LabelResize");
	RemoveBangCommand("!LabelResizeBy");
	RemoveBangCommand("!LabelScroll");
	if(IsOS(OS_2KXP))
		RemoveBangCommand("!LabelSetAlpha");
	RemoveBangCommand("!LabelSetAnimation");
	RemoveBangCommand("!LabelSetFontColor");
	RemoveBangCommand("!LabelSetFontShadowColor");
	RemoveBangCommand("!LabelSetText");
	RemoveBangCommand("!LabelShow");
	RemoveBangCommand("!LabelShowHide");
	RemoveBangCommand("!LabelToggle");
	RemoveBangCommand("!LabelGhosted");
	RemoveBangCommand("!LabelUpdate");

	RemoveBangCommand("!LabelInfoExport");

	RemoveBangCommand("!ParseEvars");
	RemoveBangCommand("!SetEvar");

	RemoveBangCommand("!PlaySound");

	#ifndef xLabelLight
	//----------------------------------------------------------------
		//xTextEdit
		if (usetextedit)
		{
			RemoveBangCommand("!xTextAppend");
			RemoveBangCommand("!xTextDelete");
			RemoveBangCommand("!xTextReplace");
			RemoveBangCommand("!xTextReplaceAll");
			RemoveBangCommand("!xTextSaveEvar");
		}

		//Shortcut2
		if (useshortcut2)
		{
			RemoveBangCommand("!ShortcutGroupHide");
			RemoveBangCommand("!ShortcutGroupShow");
			RemoveBangCommand("!ShortcutGroupToggle");
			RemoveBangCommand("!ShortcutGroupOnTop");
			RemoveBangCommand("!ShortcutGroupOnBottom");
			RemoveBangCommand("!ShortcutGroupOnTopToggle");
		}

		//xDesktop
		if (usexdesktop)
		{
			RemoveBangCommand("!xDesktopCreate");
			RemoveBangCommand("!xDesktopDestroy");
			RemoveBangCommand("!xDesktopHide");
			RemoveBangCommand("!xDesktopShow");
			RemoveBangCommand("!xDesktopToggle");
		}
	//----------------------------------------------------------------
	#endif

	LabelListIterator lit;

	for(lit = labelList.begin(); lit != labelList.end(); lit++)
	{
		#ifndef xLabelLight
		//----------------------------------------------------------------
			if ( (*lit)->labelType == 0 && !(*lit)->configSavePath.empty() )
			{
				string Search = (*lit)->getName();
				string result = "";
				
				int length = Search.length();
				int i = 0;
				
				while(i < length)
				{
					result.append(1, '[');
					result.append(1, tolower(Search[i]));
					result.append(1, toupper(Search[i++]));
					result.append(1, ']');
				}
				
				Search = result;

				result = (*lit)->configSavePath;

				//Position Saving
				string xpos = IntToString( (*lit)->getX() );
				string ypos = IntToString( (*lit)->getY() );

				string width = IntToString( (*lit)->getWidth() );
				string height = IntToString( (*lit)->getHeight() );

				string Bang = ("@" + (*lit)->configSavePath + "@ @[		]*(\\*" + Search + "[pP][oO][sS]).*@ @\\1 \"" + result + "\" \"" + xpos + "\" \"" + ypos + "\" \"" + width + "\" \"" + height + "\"@");	
				BangTextReplace(NULL, Bang.c_str());
			}
		//----------------------------------------------------------------
		#endif

		delete *lit;
	}

	labelList.clear();

	
	#ifndef xLabelLight
	//----------------------------------------------------------------
		if (usexdesktop)
		{
			for(lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
				delete *lit;

			desktoplabelList.clear();
		}

		if (useshortcut2)
		{
			for(lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
				delete *lit;

			shortcutlabelList.clear();
		}

		if (usexdesktop)
		{
			if (useautodesktop || customDelete)
			{
				if (!labelDesktopItems.empty())
				{	
					int counter = 1;

					//Clear all *xdesktopitem lines
					Bang = ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@");
					BangTextDelete(NULL, Bang.c_str());

					for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
					{
						char buffer[32];
						strcpy(buffer, "xldi");
						char counterbuffer[16];
						itoa( counter, counterbuffer, 10 );
						if (counter < 10)
							strcat(buffer, "0");
						strcat(buffer, counterbuffer);

						char name[64], data[MAX_LINE_LENGTH];
						char *linebuffers[] = {name};

						LCTokenize((*sit).c_str(), linebuffers, 1, data);

						Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + data + "@");
						BangTextAppend(NULL, Bang.c_str());

						counter++;
					}
				}

				WatchStop();
			}
		}
	//----------------------------------------------------------------
	#endif

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xLabel", hInstance);
	if(IsOS(OS_XP))
		UnregisterClass("xLabelXP", hInstance);
	UnregisterClass("xLabelMessageHandler", hInstance);

	if (defaultSettings)
		delete defaultSettings;
	if (systemInfo)
		delete systemInfo;
	
	if (hbmDesktop != 0 || hdcDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteObject(hbmDesktop);
		DeleteDC(hdcDesktop);
	}
}

Label *lookupLabel( const string &name, LabelList searchList )
{
	for(LabelListIterator lit = searchList.begin(); lit != searchList.end(); lit++)
	{
		if(stricmp(name.c_str(), (*lit)->getName().c_str()) == 0)
			return *lit;
	}

	return 0;
}

#ifndef xLabelLight
//----------------------------------------------------------------

	void RemoveDesktopItem(string path)
	{
		bool found = false;
		for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
		{
			if (stricmp((*lit)->iconCommand.c_str(), path.c_str()) == NULL)
			{
				found = true;
				break;
			}
		}
			
		if (found)
		{
			char xname[32], xx[16], xy[16], xpath[MAX_PATH_LENGTH];
			char *buffers[] = {xname, xx, xy, xpath};

			for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
			{
				LCTokenize((*sit).c_str(), buffers, 4, NULL);

				if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
				{
					labelDesktopItems.remove( *sit );
					break;
				}
			}

			(*lit)->hide();
			desktoplabelList.remove((*lit));
			delete (*lit);

			if (!useautodesktop)
				customDelete = true;
		}
	}

	UINT DirCallback( CFileInformation fiObject, EFileAction faAction, LPVOID lpData )
	{
		CString Bang;
		CString filepath = fiObject.GetFilePath();

		bool found = false;
		POSITION m_position = pathList.GetHeadPosition();
		while (m_position != NULL && found == FALSE)
		{
			if ( fiObject.IsSomeFileDir( pathList.GetAt(m_position) ) )
				found = TRUE;
			pathList.GetNext(m_position);
		}

		if (found)
		{
			if( IS_CREATE_FILE( faAction ) )
			{
				Bang = ("!xdesktopcreate updateadd \"" + filepath + "\"");
				LSExecute(NULL, Bang, SW_SHOWNORMAL);
			}
			else if( IS_DELETE_FILE( faAction ) )
			{
				Bang = ("!xdesktopcreate updatedel \"" + filepath + "\"");
				LSExecute(NULL, Bang, SW_SHOWNORMAL);
			}
			else if( IS_CHANGE_FILE( faAction ) )
			{
				Bang = ("!xdesktopcreate updateren \"" + filepath + "\"");
				LSExecute(NULL, Bang, SW_SHOWNORMAL);
			}
			else
			{
				return 1; //error, stop thread
			}
		}

		return 0; //success
	}

	void WatchStart(LPCSTR path) 
	{
		CNotifyDirCheck *DirCheck = new CNotifyDirCheck();
		
		dircheckList.insert(dircheckList.end(), DirCheck);
			
		DirCheck->SetDirectory( path );
		pathList.AddTail( path );

		//set your callback to work with each new event
		DirCheck->SetActionCallback( DirCallback );

		DirCheck->Run();
	}

	void WatchStop() 
	{
		for(DirCheckListIterator dcit = dircheckList.begin(); dcit != dircheckList.end(); dcit++)
		{
			(*dcit)->Stop();
			delete *dcit;
		}

		pathList.RemoveAll();
		dircheckList.clear();
	}

	//Read all required Files/Folder Paths (in StringList)
	StringList getActualDesktopItemPaths(bool defaultDesktop, string searchPath)
	{
		StringList ActualDesktopItemPaths;

		string wildcardPath = "";
		string itemName = "";
		
		char tmp[MAX_PATH_LENGTH];
		string tmpHelper = "";

		HANDLE hFindHandle;
		WIN32_FIND_DATA FindData;

		if (defaultDesktop)
		{
			//Create 4 Main ShellFolders: Workspace, My Documents, NetworkHood, Trash 
			if ( !GetxRCBool("xdesktop", "NoWorkSpace") )
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}");
			if ( !GetxRCBool("xdesktop", "NoMyDocuments") )
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}");
			if ( !GetxRCBool("xdesktop", "NoNetwork") )
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}");
			if ( !GetxRCBool("xdesktop", "NoTrashcan") )
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{645FF040-5081-101B-9F08-00AA002F954E}");

			VarExpansion( tmp, "$commondesktopdir$" );
			searchPath = tmp;
			wildcardPath = (searchPath + "*.*");

			hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

			for(;;)
			{
				if(INVALID_HANDLE_VALUE == hFindHandle)
					break;

				itemName = FindData.cFileName;

				//Make sure its a valid folder
				if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
				{
					tmpHelper = (searchPath + itemName);
					ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
				}

				if(!FindNextFile(hFindHandle, &FindData))
					break;
			}

			VarExpansion( tmp, "$desktopdir$" );
			searchPath = tmp;
			wildcardPath = (searchPath + "*.*");

			hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

			for(;;)
			{
				if(INVALID_HANDLE_VALUE == hFindHandle)
					break;
						
				itemName = FindData.cFileName;

				//Make sure its a valid folder
				if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
				{
					tmpHelper = (searchPath + itemName);
					ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
				}

				if(!FindNextFile(hFindHandle, &FindData))
					break;
			}
		}
		else
		{
			//Check for ending BackSlash
			tmpHelper = searchPath.substr(searchPath.length()-1, searchPath.length());

			if (tmpHelper != "\\")
				wildcardPath = (searchPath + "\\*.*");
			else
				wildcardPath = (searchPath + "*.*");

			hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

			for(;;)
			{
				if(INVALID_HANDLE_VALUE == hFindHandle)
					break;
						
				itemName = FindData.cFileName;

				//Make sure its a valid folder
				if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
				{
					tmpHelper = (searchPath + itemName);
					ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
				}

				if(!FindNextFile(hFindHandle, &FindData))
					break;
			}
		}

		FindClose(hFindHandle);

		return ActualDesktopItemPaths;
	}

	//Write all Items to File
	void DesktopUpdate(int updateMode, string updatedFile, StringList ActualDesktopItemPaths)
	{
		int counter = 1;

		//Saves already present Items, Only if:
		//	1. They exist
		//  2. It's wished
		//  3. AutoCreateDesktop isn't running
		if (!labelDesktopItems.empty() && updateMode == 1 && !useautodesktop)
		{
			for(StringListIterator it = labelDesktopItems.begin(); it != labelDesktopItems.end(); it++)
			{
				char buffer[32];
				strcpy(buffer, "xldi");
				char counterbuffer[16];
				itoa( counter, counterbuffer, 10 );
				if (counter < 10)
					strcat(buffer, "0");
				strcat(buffer, counterbuffer);

				char name[64], data[MAX_LINE_LENGTH];
				char *linebuffers[] = {name};

				LCTokenize((*it).c_str(), linebuffers, 1, data);

				Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + data + "@");
				BangTextAppend(NULL, Bang.c_str());

				counter++;
			}
		}
		
		//Saves new Items, Only if:
		//	1. It's wished (Replace)
		//  2. It's wished (Append)
		//  3. None exists (Initial Creation required)
		//  AND IT'S NOT AN UPDATE
		if ( (updateMode == 0 || updateMode == 1 || labelDesktopItems.empty()) && updateMode != -1)
		{
			int tempcounter = 1;

			NameValuePair directionValues[] = {
			{ "left", 1 },
			{ "right", 2 },
			{ "up", 3 },
			{ "down", 4 },
			{ 0, 0 }
			};

			NameValuePair wrapdirectionValues[] = {
			{ "up", 1 },
			{ "down", 2 },
			{ "left", 3 },
			{ "right", 4 },
			{ 0, 0 }
			};

			int DIwidth = GetxRCDimen("xdesktop", "ItemWidth", 100, GetSystemMetrics(SM_CXSCREEN));
			int DIheight = GetxRCDimen("xdesktop", "ItemHeight", 75, GetSystemMetrics(SM_CYSCREEN));
			int xSpace = GetxRCInt("xdesktop", "SpacingX", 0);
			int ySpace = GetxRCInt("xdesktop", "SpacingY", 0);
					
			int direction = GetxRCNamedValue("xdesktop", "Direction", directionValues, 2);
			int wrapDirection = GetxRCNamedValue("xdesktop", "WrapDirection", wrapdirectionValues, 2);
			
			int wrapCount;
			if (wrapDirection == 1 || wrapDirection == 2)
				wrapCount = GetxRCInt("xdesktop", "WrapCount", (GetSystemMetrics(SM_CXSCREEN)-xSpace)/(DIwidth+xSpace));
			else if (wrapDirection == 3 || wrapDirection == 4)
				wrapCount = GetxRCInt("xdesktop", "WrapCount", (GetSystemMetrics(SM_CYSCREEN)-ySpace)/(DIheight+ySpace));

			int DIStartx = GetxRCCoord("xdesktop", "X", ((GetSystemMetrics(SM_CXSCREEN)-(wrapCount*(DIwidth+xSpace)-xSpace))/2 < 0 ? 0 : (GetSystemMetrics(SM_CXSCREEN)-(wrapCount*(DIwidth+xSpace)-xSpace))/2), GetSystemMetrics(SM_CXSCREEN));
			int DIStarty = GetxRCCoord("xdesktop", "Y", ((GetSystemMetrics(SM_CYSCREEN)-(wrapCount*(DIheight+ySpace)-ySpace))/2 < 50 ? 50 : (GetSystemMetrics(SM_CYSCREEN)-(wrapCount*(DIheight+ySpace)-ySpace))/2), GetSystemMetrics(SM_CYSCREEN));

			int tempx;
			int tempy;

			if (direction == 1)
				tempx = DIStartx+DIwidth+xSpace;
			else if (direction == 2)
				tempx = DIStartx-(DIwidth+xSpace);
			else if (direction == 3)
				tempy = DIStarty+DIheight+ySpace;
			else if (direction == 4)
				tempy = DIStarty-(DIheight+ySpace);

			if (direction == 1 || direction == 2)
				tempy = DIStarty;
			else if (direction == 3 || direction == 4)
				tempx = DIStartx;

			for(StringListIterator it = ActualDesktopItemPaths.begin(); it != ActualDesktopItemPaths.end(); it++)
			{
				char buffer[32];
				strcpy(buffer, "xldi");
				char counterbuffer[16];
				char xbuffer[16];
				char ybuffer[16];
				itoa( counter, counterbuffer, 10 );
				if (counter < 10)
					strcat(buffer, "0");
				strcat(buffer, counterbuffer);

				//Sorting/Positioning due to settings
				div_t div_result = div( tempcounter, wrapCount );

				if (direction == 1 || direction == 2)
				{
					if (wrapDirection > 2)
						wrapDirection = 2;
							
					if (div_result.quot > 0 && div_result.rem == 1)
					{
						tempx = DIStartx;
						if (wrapDirection == 1)
							tempy = DIStarty-(div_result.quot*(DIheight+ySpace));
						else if (wrapDirection == 2)
							tempy = DIStarty+div_result.quot*(DIheight+ySpace);
					}
					else
					{
						if (direction == 1)
							tempx = tempx-(DIwidth+xSpace);
						else if (direction == 2)
							tempx = tempx+DIwidth+xSpace;
					}
				}
				if (direction == 3 || direction == 4)
				{
					if (wrapDirection < 3)
						wrapDirection = 4;
						
					if (div_result.quot > 0 && div_result.rem == 1)
					{
						tempy = DIStarty;
						if (wrapDirection == 3)
							tempx = DIStartx-(div_result.quot*(DIwidth+xSpace));
						else if (wrapDirection == 4)
							tempx = DIStartx+div_result.quot*(DIwidth+xSpace);
					}
					else
					{
						if (direction == 3)
							tempy = tempy-(DIheight+ySpace);
						else if (direction == 4)
							tempy = tempy+DIheight+ySpace;
					}
				}

				itoa( tempx, xbuffer, 10 );
				itoa( tempy, ybuffer, 10 );

				if (!useautodesktop)
				{
					Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + xbuffer + " " + ybuffer + " \"" + (*it) + "\"@");
					BangTextAppend(NULL, Bang.c_str());
				}

				Label *label = new Label(buffer);

				desktoplabelList.insert(desktoplabelList.end(), label);

				Bang = "dummy ";
				Bang.append( xbuffer );
				Bang.append( " ");
				Bang.append( ybuffer );
				Bang.append( " \"" + (*it) + "\"");

				label->desktopload(hInstance, Bang.c_str());

				labelDesktopItems.insert(labelDesktopItems.end(), Bang);

				counter++;
				tempcounter++;
			}
		}
		//Update & Saves new Items, Only if:
		//	1. Some have changed
		else if (updateMode == -1)
		{
			for (StringListIterator sit = ActualDesktopItemPaths.begin(); sit != ActualDesktopItemPaths.end(); sit++)
			{
				bool found = false;

				for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
				{
					if (stricmp((*sit).c_str(), (*lit)->iconCommand.c_str()) == NULL)
					{
						found = true;
						break;
					}
				}

				if (found)
				{
					int left = (*lit)->getX();
					int top = (*lit)->getY();

					if (left < 0)
						left = 0;
					if (top < 0)
						top = 0;
								
					string position = ( IntToString(left) + " " + IntToString(top) );

					Bang = ( "dummy " + position + " \"" + (*lit)->iconCommand + "\"");

					labelDesktopItems.insert(labelDesktopItems.end(), Bang);

					(*lit)->hide();
					desktoplabelList.remove((*lit));
					delete (*lit);
				}
				else
				{
					Bang = ( "dummy 100 100 \"" + (*sit) + "\"" );
				
					labelDesktopItems.insert(labelDesktopItems.end(), Bang);
				}
			}
			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				(*lit)->hide();
				desktoplabelList.remove((*lit));
				delete (*lit);
			}

			desktoplabelList.clear();
			
			int counter = 1;
			for(sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
			{
				char buffer[32];
				strcpy(buffer, "xldi");
				char counterbuffer[16];
				itoa( counter, counterbuffer, 10 );
				if (counter < 10)
					strcat(buffer, "0");
				strcat(buffer, counterbuffer);

				char name[64], data[MAX_LINE_LENGTH];
				char *linebuffers[] = {name};

				LCTokenize((*sit).c_str(), linebuffers, 1, data);

				Label *label = new Label(name);

				desktoplabelList.insert(desktoplabelList.end(), label);

				label->desktopload(hInstance, (*sit));

				if (!useautodesktop)
				{
					Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + data + "@");
					BangTextAppend(NULL, Bang.c_str());
				}
				counter++;
			}
		}
		//Update & Saves single changed Items, Only if:
		//	1. they have changed
		//	1. AutoCreateDesktop is ON
		else if (updateMode >= 20 && useautodesktop)
		{
			//Add Files
			if (updateMode == 21)
			{
				Bang = ( "dummy 100 100 \"" + updatedFile + "\"" );
			
				labelDesktopItems.insert(labelDesktopItems.end(), Bang);

				Label *label = new Label("dummy");

				desktoplabelList.insert(desktoplabelList.end(), label);

				label->desktopload(hInstance, Bang.c_str());
			}
			//Remove Files
			else if (updateMode == 22)
			{
				bool found = false;
				for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
				{
					if (stricmp((*lit)->iconCommand.c_str(), updatedFile.c_str()) == NULL)
					{
						found = true;
						break;
					}
				}
				
				if (found)
				{
					char xname[32], xx[16], xy[16], xpath[MAX_PATH_LENGTH];
					char *buffers[] = {xname, xx, xy, xpath};

					for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
					{
						LCTokenize((*sit).c_str(), buffers, 4, NULL);

						if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
						{
							labelDesktopItems.remove( *sit );
							break;
						}
					}

					(*lit)->hide();
					desktoplabelList.remove((*lit));
					delete (*lit);
				}
			}
			//Rename Files
			else if (updateMode == 23)
			{
				HANDLE hFindHandle;
				WIN32_FIND_DATA FindData;

				for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
				{
					if (strstr((*lit)->iconCommand.c_str(), "::") == NULL)
					{
						hFindHandle = FindFirstFile((*lit)->iconCommand.c_str(), &FindData);

						if(INVALID_HANDLE_VALUE == hFindHandle)
						{
							char xname[32], xx[16], xy[16], xpath[MAX_PATH_LENGTH];
							char *buffers[] = {xname, xx, xy, xpath};

							for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
							{
								LCTokenize((*sit).c_str(), buffers, 4, NULL);

								if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
								{
									labelDesktopItems.remove( *sit );
									break;
								}
							}

							string positions = ( IntToString((*lit)->getX()) + " " + IntToString((*lit)->getY()) );

							(*lit)->hide();
							delete *lit;
							desktoplabelList.remove( *lit );

							Bang = ( "dummy " + positions + " \"" + updatedFile + "\"" );

							//New Named Label
							labelDesktopItems.insert(labelDesktopItems.end(), Bang);

							Label *label = new Label("dummy");

							desktoplabelList.insert(desktoplabelList.end(), label);

							label->desktopload(hInstance, Bang.c_str());

							break;
						}
					}
				}

				FindClose(hFindHandle);
			}
		}
	}

	// !xDesktopCreate <append>|<replace>|<update...> <path to lnkfolder>|<updatefile>
	void BangDesktopCreate(HWND hwndCaller, LPCSTR pszArgs)
	{
		StringList labelDesktopItemPaths;

		char action[32], path[MAX_LINE_LENGTH];
		char *buffers[] = {action, path};

		LCTokenize(pszArgs, buffers, 2, NULL);
		
		string searchPath = path;

		//Quit here, if no path is specified
		if (searchPath.empty())
			return;

		//Initialize
		bool defaultDesktop = false;
		
		if ( stricmp(searchPath.c_str(), ".desktop") == 0 || stricmp(searchPath.c_str(), "deskto") == 0 || stricmp(searchPath.c_str(), "eskto") == 0 )
			defaultDesktop = true;
		
		//Start Configuration and Writing to File
		if (!desktopSavePath.empty())
		{
			int updateMode;
			if ( stricmp(action, "update") == NULL )
				updateMode = -1;
			else if ( stricmp(action, "updateadd") == NULL )
				updateMode = 21;
			else if ( stricmp(action, "updatedel") == NULL )
				updateMode = 22;
			else if ( stricmp(action, "updateren") == NULL )
				updateMode = 23;
			else if (stricmp(action, "append") == NULL)
				updateMode = 1;
			else 
				updateMode = 0;

			if (updateMode == -1 || updateMode == 0 || updateMode == 1)
			{
				//Get All Items actual items
				labelDesktopItemPaths = getActualDesktopItemPaths(defaultDesktop, searchPath);

				//Clear all *xdesktopitem lines
				Bang = ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@");
				BangTextDelete(NULL, Bang.c_str());

				if (updateMode == -1 || updateMode == 0)
				{
					labelDesktopItems.clear();

					if (updateMode == 0)
						BangDestroy(NULL, "xdesktop");
				}
			}

			//0 = Create New (Replace/Reposition), 1 = Append (New will be Positioned), 2+ = Update
			DesktopUpdate(updateMode, searchPath, labelDesktopItemPaths);
		}
		else
		{
			Bang = "!alert \"You haven't defined the setting\nxDesktopConfigFile\nwhich points to the file containing your *xDesktopItem lines!\" \"xLabel Debug Information\"";
			LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
		}
	}

	void InternalSave(string internalcaller)
	{
		if (!desktoplabelList.empty() && internalcaller.empty() && !useautodesktop)
		{
			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				int left = (*lit)->getX();
				int top = (*lit)->getY();

				if (left < 0)
					left = 0;
				if (top < 0)
					top = 0;
							
				string position = ( IntToString(left) + " " + IntToString(top) );

				Bang = ("@" + (*lit)->configSavePath + "@ @[		]*(\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]).*(" + (*lit)->getName() + ")@ @\\1 \\2 " + position + " \"" + (*lit)->iconCommand + "\"@");
				if (!(*lit)->configSavePath.empty())
					BangTextReplace(NULL, Bang.c_str());
				else
				{
					Bang = "!alert \"You haven't defined the setting\nxDesktopConfigFile\nwhich points to the file containing your *xDesktopItem lines!\" \"xLabel Debug Information\"";
					LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
				}
			}
			//Stop if complete Desktop was saved
			return;
		}
		else
		{
			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
			{
				if (stricmp((*lit)->iconCommand.c_str(), internalcaller.c_str()) == NULL)
				{
					int left = (*lit)->getX();
					int top = (*lit)->getY();

					if (left < 0)
						left = 0;
					if (top < 0)
						top = 0;
							
					string position = ( IntToString(left) + " " + IntToString(top) );

					//Only Track change and do not save, if AutoCreate Desktop
					if (useautodesktop)
					{
						char xname[32], xx[16], xy[16], xpath[MAX_PATH_LENGTH];
						char *buffers[] = {xname, xx, xy, xpath};

						for(StringListIterator sit = labelDesktopItems.begin(); sit != labelDesktopItems.end(); sit++)
						{
							LCTokenize((*sit).c_str(), buffers, 4, NULL);

							if (stricmp((*lit)->iconCommand.c_str(), xpath) == NULL)
							{
								Bang = ( (*lit)->getName() + " " + position + " \"" + (*lit)->iconCommand + "\"" );
								
								*sit = Bang;
								
								break;
							}
						}
					}
					//Save immediately, if no AutoCreate Desktop
					else
					{
						Bang = ("@" + (*lit)->configSavePath + "@ @[		]*(\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]).*(" + (*lit)->getName() + ")@ @\\1 \\2 " + position + " \"" + (*lit)->iconCommand + "\"@");
						if (!(*lit)->configSavePath.empty())
							BangTextReplace(NULL, Bang.c_str());
					}

					return;
				}
			}
		}
		//No Complete Desktop Save and No Single DesktopItem Save
		//Try Shortcuts ;) 
		Label *label = lookupLabel(internalcaller.c_str(), shortcutlabelList);

		if(label)
		{
			int left = label->getX();
			int top = label->getY();

			if (left < 0)
				left = 0;
			if (top < 0)
				top = 0;
						
			string position = ( IntToString(left) + " " + IntToString(top) );
			
			Bang = ("@" + label->configSavePath + "@ @[		]*(\\*[sS][hH][oO][rR][tT][cC][uU][tT])([ 	]*)(\\=" + label->getName() + "\\=)([ 	]*)([^ 	]+)([ 	]*)([^ 	]+)([ 	]*)(.*)@ @\\1 \\3 " + position + " \\9@");
			if (!label->configSavePath.empty())
				BangTextReplace(NULL, Bang.c_str());
			else
			{
				Bang = "!alert \"You haven't defined the setting\nShortcutConfigFile\nwhich points to the file containing your *shortcut lines!\" \"xLabel Debug Information\"";
				LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
			}
		}
	}

	void BangDesktopDestroy(HWND hwndCaller, LPCSTR pszArgs)
	{
		BangDestroy(NULL, "xdesktop");
	}

	void BangDesktopHide(HWND hwndCaller, LPCSTR pszArgs)
	{
		BangHide(NULL, "xdesktop");
	}

	void BangDesktopShow(HWND hwndCaller, LPCSTR pszArgs)
	{
		BangShow(NULL, "xdesktop");
	}

	void BangDesktopToggle(HWND hwndCaller, LPCSTR pszArgs)
	{
		BangToggle(NULL, "xdesktop");
	}
//----------------------------------------------------------------
#endif

// !LabelInfoExport < clipboard | mzscript var | $Evar$> <Text Escape Sequence>
void BangInfoExport(HWND hwndCaller, LPCSTR pszArgs)
{
	char namevar[64], textvar[MAX_PATH_LENGTH];
	char *buffers[] = {namevar, textvar};

	if (LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{	
		Label *label = lookupLabel(namevar);

		if(label)
		{
			string evar = textvar;

			if (strstr(evar.c_str(), "%#") != NULL)
			{
				while ( strstr(evar.c_str(), "%#") != NULL )
				{
					evar.replace(evar.find("%#"), 2, "", 0);
				}

				Bang = ("\"" + label->getText() + "\"");
				PrefixedLSSetVariable( evar, "", Bang );
			}
			else
			{
				Bang = evar;
				Bang.append(" \"" + label->getText() + "\"");
				ParseBangCommand(NULL, "!varset", Bang.c_str());
			}
		}
		else
		{
			string result = "";
			string evar = namevar;
			bool dynamic = false;
			result = systemInfo->processLabelText(textvar, (*labelList.begin()), &dynamic);

			if ( stricmp(evar.c_str(), "clipboard") == NULL)
			{
				SetClipboardText(NULL, result.c_str());
			}
			else if (strstr(evar.c_str(), "%#") != NULL)
			{
				while ( strstr(evar.c_str(), "%#") != NULL )
				{
					evar.replace(evar.find("%#"), 2, "", 0);
				}

				Bang = ("\"" + result + "\"");
				PrefixedLSSetVariable( evar, "", Bang );
			}
			else
			{
				Bang = evar;
				Bang.append(" \"" + result + "\"");
				ParseBangCommand(NULL, "!varset", Bang.c_str());
			}
		}
	}
}

// !LabelSetAnimation <label> <on | off> [<loops>]
void BangSetAnimation(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16], loops[16];
	char *buffers[] = {name, option, loops};

	int numTokens = LCTokenize(pszArgs, buffers, 3, NULL);
	if (numTokens >= 2)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (numTokens == 2)
			{
				if(stricmp(option, "off") == 0)
					label->setAnimation(false);
				else if(stricmp(option, "on") == 0)
					label->setAnimation(true);
			}
			else if  (numTokens == 3)
			{
				if(stricmp(option, "off") == 0)
					label->setAnimation(false);
				else if(stricmp(option, "on") == 0)
					label->setAnimation(true, atoi(loops));
			}
		}
	}
}

// !LabelAlwaysOnTop <label> <on|true, off|false, toggle>
void BangAlwaysOnTop(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[16];
	char *buffers[] = {name, option};

	LCTokenize(pszArgs, buffers, 2, NULL);

	Label *label = lookupLabel(name);
	if(label)
	{
		if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
			label->setAlwaysOnTop(false);
		else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
			label->setAlwaysOnTop(!label->isAlwaysOnTop());
		else
			label->setAlwaysOnTop(true);
	}
	else
	{
		LabelListIterator lit;
		for( lit = labelList.begin(); lit != labelList.end(); lit++ )
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
			{
				if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
					(*lit)->setAlwaysOnTop(false);
				else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
					(*lit)->setAlwaysOnTop(!(*lit)->isAlwaysOnTop());
				else
					(*lit)->setAlwaysOnTop(true);
			}
		}

		char group[MAX_LINE_LENGTH];
		LPCSTR nextGroup = pszArgs;

		while (GetToken(nextGroup, group, &nextGroup, false))
		{
			for( lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), group) == 0 || stricmp((*lit)->labelGroup.c_str(), "*") == 0 )
					(*lit)->setAlwaysOnTop(true);
			}
		}
	}
}

// !LabelClipboardCopy <label> <prefix>
void BangClipboardCopy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardCopy((numTokens >= 2) ? prefix : "");
	}
}

// !LabelClipboardPaste <label> <prefix>
void BangClipboardPaste(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], prefix[256];
	char *buffers[] = {name, prefix};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->clipboardPaste((numTokens >= 2) ? prefix : "");
	}
}

// !LabelCreate <label>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = new Label(pszArgs);
	labelList.insert(labelList.end(), label);

	if ( IsOS(OS_XP) && GetxRCBool(pszArgs, "DropShadow") )
		label->load(hInstance, 0, true);
	else 
		label->load(hInstance);

	//ModuleHook
	if (!label->moduleHook.empty())
	{
		for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
		{
			string hWndbuffer = IntToString( (int) label->gethWnd() );
			string config = (*sit).c_str();

			if ( !label->configSavePath.empty() )
			{
				config.append( " \"" );
				config.append( label->configSavePath.c_str() );
				config.append( "\" " );
				config.append( hWndbuffer );
			}
			else
			{
				config.append( " \"$themedir$theme.rc\" " );
				config.append( hWndbuffer );
			}

			LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
		}
	}

	if (!label->isStartHidden())
		label->show();
	else
		label->hide();

	//OnResize one time after all are loaded and shown
	//if( !label->resizeCommand.empty() )
	//	xLSExecute(label->gethWnd(), label->resizeCommand.c_str(), SW_SHOWNORMAL);
}

// !LabelDestroy <label>
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{
	#ifndef xLabelLight
	//----------------------------------------------------------------
		if (stricmp(pszArgs, "xdesktop") == 0)
		{
			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
				delete *lit;

			desktoplabelList.clear();
		}
		else
		{
	//----------------------------------------------------------------
	#endif
		Label *label = lookupLabel(pszArgs);
		if(label)
		{
			label->hide();
			labelList.remove(label);
			delete label;
		}
		else
		{
			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
				{
					(*lit)->hide();
					labelList.remove((*lit));
					delete (*lit);
				}
			}
		}
	#ifndef xLabelLight
	//----------------------------------------------------------------
		}
	//----------------------------------------------------------------
	#endif
}

// !LabelDock <label> <label> <"true", "false">
void BangDock(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], dockto[64], state[32];
	char *buffers[] = {name, dockto, state};

	if(LCTokenize(pszArgs, buffers, 3, NULL) >= 2)
	{	
		Label *label = lookupLabel(name);
		Label *labeldockto = lookupLabel(dockto);

		if (stricmp( state, "false" ) == NULL)
		{
			if(label && labeldockto)
			{
				for(LabelListIterator lit = labeldockto->hasDockedLabelList.begin(); lit != labeldockto->hasDockedLabelList.end(); lit++)
				{
					if ( stricmp( name, (*lit)->getName().c_str() ) == NULL )
					{
						labeldockto->hasDockedLabelList.erase ( lit );
						break;
					}
				}
				
			}
		}
		else
		{
			if(label && labeldockto)
			{
				bool alreadyFound = false;
				for(LabelListIterator lit = labeldockto->hasDockedLabelList.begin(); lit != labeldockto->hasDockedLabelList.end(); lit++)
				{
					if ( stricmp( name, (*lit)->getName().c_str() ) == NULL )
					{
						alreadyFound = true;
						break;
					}
				}
				if ( !alreadyFound )
					labeldockto->hasDockedLabelList.insert ( labeldockto->hasDockedLabelList.end(), label );
			}
		}
	}
}

// !LabelShow <label>
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
	#ifndef xLabelLight
	//----------------------------------------------------------------
		if (stricmp(pszArgs, "xdesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				(*i)->show();
		}
		else
		{
	//----------------------------------------------------------------
	#endif
		Label *label = lookupLabel(pszArgs);
		if(label)
			label->show(true);
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
				{
					(*lit)->show(true);
				}
			}
		
			#ifndef xLabelLight
			//----------------------------------------------------------------
				char group[MAX_LINE_LENGTH];
				LPCSTR nextGroup = pszArgs;

				while (GetToken(nextGroup, group, &nextGroup, false))
				{
					for( lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
					{
						if( stricmp((*lit)->labelGroup.c_str(), group) == 0 || stricmp((*lit)->labelGroup.c_str(), "*") == 0 )
							(*lit)->show();
					}
				}
			//----------------------------------------------------------------
			#endif
		}
	#ifndef xLabelLight
	//----------------------------------------------------------------
		}
	//----------------------------------------------------------------
	#endif
}

// !LabelShowHide <label> <timeout>
void BangShowHide(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[64];
	char *buffers[] = {name, A};

	if(LCTokenize(pszArgs, buffers, 2, NULL) >= 2)
	{
		int timeout = atoi(A);

		#ifndef xLabelLight
		//----------------------------------------------------------------
			if (stricmp(name, "xdesktop") == 0)
			{
				for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
					(*i)->showHide(timeout);
			}
			else
			{
		//----------------------------------------------------------------
		#endif
			Label *label = lookupLabel(name);

			if(label)
				label->showHide(timeout);
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
						(*i)->showHide(timeout);
				}
			}
		#ifndef xLabelLight
		//----------------------------------------------------------------
			}
		//----------------------------------------------------------------
		#endif
	}
}

// !LabelHide <label>
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
	#ifndef xLabelLight
	//----------------------------------------------------------------
		if (stricmp(pszArgs, "xdesktop") == 0)
		{
			for(LabelListIterator lit = desktoplabelList.begin(); lit != desktoplabelList.end(); lit++)
				(*lit)->hide();
		}
		else
		{
	//----------------------------------------------------------------
	#endif
		Label *label = lookupLabel(pszArgs);
		if(label)
			label->hide(true);
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++ )
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
				{
					(*lit)->hide(true);
				}
			}

			#ifndef xLabelLight
			//----------------------------------------------------------------
				char group[MAX_LINE_LENGTH];
				LPCSTR nextGroup = pszArgs;

				while (GetToken(nextGroup, group, &nextGroup, false))
				{
					for( lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
					{
						if( stricmp((*lit)->labelGroup.c_str(), group) == 0 || stricmp((*lit)->labelGroup.c_str(), "*") == 0 )
							(*lit)->hide();
					}
				}
			//----------------------------------------------------------------
			#endif
		}
	#ifndef xLabelLight
	//----------------------------------------------------------------		
		}
	//----------------------------------------------------------------
	#endif
}

// !LabelLSBoxHook <labelname>
void BangLSBoxHook(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], rest[MAX_PATH_LENGTH];
	char *buffers[] = {name};

	int tokens = LCTokenize(pszArgs, buffers, 1, rest);

	const char *handle = strrchr(pszArgs,' ');

	HWND hwndBox = (HWND)atoi(handle+1);

	Label *label = lookupLabel(name);

	if(label)
	{
		label->setBox(hwndBox);
		label->update();		
	}
	else
	{	
		#ifndef xLabelLight
		//----------------------------------------------------------------
			if ( strstr(rest, "#") != NULL && (strstr(rest, "!") != NULL || strstr(rest, "[!") != NULL) )
			{
				label = new Label(name);
				shortcutlabelList.insert(shortcutlabelList.end(), label);
				label->shortcutload(hInstance, pszArgs, 0, 0, hwndBox);

				return;
			}
		//----------------------------------------------------------------
		#endif

		label = new Label(name);
		labelList.insert(labelList.end(), label);
		label->load(hInstance, hwndBox);

		//ModuleHook
		if (!label->moduleHook.empty())
		{
			for(StringListIterator sit = label->moduleHook.begin(); sit != label->moduleHook.end(); sit++)
			{
				string hWndbuffer = IntToString( (int) label->gethWnd() );
				string config = (*sit).c_str();

				if ( !label->configSavePath.empty() )
				{
					config.append( " \"" );
					config.append( label->configSavePath.c_str() );
					config.append( "\" " );
					config.append( hWndbuffer );
				}
				else
				{
					config.append( " \"$themedir$theme.rc\" " );
					config.append( hWndbuffer );
				}

				LSExecute(NULL, config.c_str(), SW_SHOWNORMAL);
			}
		}

		if (!label->isStartHidden())
			label->show();
		else
			label->hide();

		//OnResize one time after all are loaded and shown
		//if( !label->resizeCommand.empty() )
		//	xLSExecute(label->gethWnd(), label->resizeCommand.c_str(), SW_SHOWNORMAL);
	}
}

// !LabelMove <label> <x> <y> [<steps> <time>]
void BangMove(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if(LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, GetSystemMetrics(SM_CXSCREEN));
				y = xParseCoord(B, 0, GetSystemMetrics(SM_CYSCREEN));
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right-rc.left);
				y = xParseCoord(B, 0, rc.bottom-rc.top);
			}
			int step = atoi( C );
			int time = atoi( D );

			label->reposition(1, x, y, 0, 0, step, time);
		}
	}
}

// !LabelMoveBy <label> <x> <y> [<steps> <time>] [bool s]
void BangMoveBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if(LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int x = atoi( A );
		int y = atoi( B );

		int step = atoi( C );
		int time = atoi( D );

		if(label)
			label->reposition(1, x, y, 0, 0, step, time, true);
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), name) == NULL )
					(*lit)->reposition(1, x, y, 0, 0, step, time, true);
			}
		}
	}
}

// !LabelNext <label>
void BangNext(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->next();
}

// !LabelPrevious <label>
void BangPrevious(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->previous();
}

// !LabelRefresh <label>
void BangRefresh(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], setting[256], value[MAX_PATH_LENGTH];
	char *buffers[] = {name, setting};

	int tokens = LCTokenize(pszArgs, buffers, 2, value);

	Label *label = lookupLabel(name);

	if(label)
	{
		if (tokens == 1)
		{
			label->reconfigure(true);
			label->repaint(true);
		}

		if (tokens == 2)
		{
			string updatevalue = value;
			//Real Negative Fix

			if (stricmp(setting, "x") == NULL || stricmp(setting, "y") == NULL)
			{
				updatevalue.replace(updatevalue.find("-"), 1, "~", 1);
			}

			PrefixedLSSetVariable( (string)setting, "", updatevalue );
			label->reconfigure(true);
			label->repaint(true);
		}
	}
	else
	{
		LabelListIterator lit;
		for( lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
			{
				if (tokens == 1)
				{
					label->reconfigure(true);
					label->repaint(true);
				}

				if (tokens == 2)
				{
					string updatevalue = value;
					//Real Negative Fix
					
					if (stricmp(setting, "x") == NULL || stricmp(setting, "y") == NULL)
					{
						updatevalue.replace(updatevalue.find("-"), 1, "~", 1);
					}

					string worker = setting;

					if (strstr(setting, "*") != NULL)
						worker.replace(worker.find("*"), 1, (*lit)->getName().c_str(), (*lit)->getName().length());

					PrefixedLSSetVariable( worker, "", (string)value );
					(*lit)->reconfigure(true);
					(*lit)->repaint(true);
				}
			}
		}
	}
}

// !LabelReposition <label> <x> <y> <width> <height> [<steps>] [<time>]
void BangReposition(HWND hwndCaller, LPCSTR pszArgs)
{
	int x = 0;
	int y = 0;
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if(LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (label->getBox() == NULL)
			{
				x = xParseCoord(A, 0, GetSystemMetrics(SM_CXSCREEN));
				y = xParseCoord(B, 0, GetSystemMetrics(SM_CYSCREEN));
				w = xParseDimen(C, 0, GetSystemMetrics(SM_CXSCREEN));
				h = xParseDimen(D, 0, GetSystemMetrics(SM_CYSCREEN));
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				x = xParseCoord(A, 0, rc.right-rc.left);
				y = xParseCoord(B, 0, rc.bottom-rc.top);
				w = xParseDimen(C, 0, rc.right-rc.left);
				h = xParseDimen(D, 0, rc.bottom-rc.top);
			}
			int step = atoi( E );
			int time = atoi( F );

			label->reposition(3, x, y, w, h, step, time);
		}
	}
}

// !LabelRepositionBy <label> <cx> <cy> <cwidth> <cheight> [<steps>] [<time>]
void BangRepositionBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16], E[16], F[16];
	char *buffers[] = {name, A, B, C, D, E, F};

	if(LCTokenize(pszArgs, buffers, 7, NULL) >= 5)
	{
		Label *label = lookupLabel(name);
		
		int x = atoi( A );
		int y = atoi( B );
		int w = atoi( C );
		int h = atoi( D );

		int step = atoi( E );
		int time = atoi( F );

		if(label)
			label->reposition(3, x, y, w, h, step, time, true);
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
					(*lit)->reposition(3, x, y, w, h, step, time, true);
			}
		}
	}
}


// !LabelResize <label> <width> <height> [<steps> <time>]
void BangResize(HWND hwndCaller, LPCSTR pszArgs)
{
	int w = 64;
	int h = 64;
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if(LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if (label->getBox() == NULL)
			{
				w = xParseDimen(A, 0, GetSystemMetrics(SM_CXSCREEN));
				h = xParseDimen(B, 0, GetSystemMetrics(SM_CYSCREEN));
			}
			else
			{
				RECT rc;
				GetClientRect(label->getBox(), &rc);
				w = xParseDimen(A, 0, rc.right-rc.left);
				h = xParseDimen(B, 0, rc.bottom-rc.top);
			}
			int step = atoi( C );
			int time = atoi( D );

			label->reposition(2, 0, 0, w, h, step, time);
		}
	}
}

// !LabelResizeBy <label> <cx> <cy> [<steps> <time>]
void BangResizeBy(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16], D[16];
	char *buffers[] = {name, A, B, C, D};

	if(LCTokenize(pszArgs, buffers, 5, NULL) >= 3)
	{
		Label *label = lookupLabel(name);

		int cx = atoi( A );
		int cy = atoi( B );

		int step = atoi( C );
		int time = atoi( D );

		if(label)
			label->reposition(2, 0, 0, cx, cy, step, time, true);
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
					(*lit)->reposition(2, 0, 0, cx, cy, step, time, true);
			}
		}
	}
}

// !LabelScroll <label> <'on', 'off', 'toggle', 'freeze', limit> <scrollMode(1-4)>
void BangScroll(HWND hwndCaller, LPCSTR pszArgs)
{
	// FIXME: This should be split into multiple bang commands
	char name[64], option[16], mode[16];
	char *buffers[] = {name, option, mode};

	if(LCTokenize(pszArgs, buffers, 3, NULL) >= 2)
	{
		Label *label = lookupLabel(name);

		if(label)
		{
			if(stricmp(option, "off") == 0)
			{
				label->setScrolling(0);
				if (label->frozen)
					label->frozen = false;
			}
			else if(stricmp(option, "on") == 0)
			{
				int scrollmode = atoi(mode);
				if (scrollmode > 0 && scrollmode < 5)
					label->setScrolling(scrollmode);
				else
					label->setScrolling(label->scrollBackup);
			}
			else if(stricmp(option, "toggle") == 0)
			{
				if (label->getScrolling() > 0)
				{
					label->setScrolling(0);
					if (label->frozen)
						label->frozen = false;
				}
				else
				{
					int scrollmode = atoi(mode);
					if (scrollmode > 0 && scrollmode < 5)
						label->setScrolling(scrollmode);
					else
						label->setScrolling(label->scrollBackup);
				}
			}
			else if(stricmp(option, "freeze") == 0)
			{
				if (label->frozen)
					label->frozen = false;
				else
					label->frozen = true;
			}
			else
			{
				int limit = atoi(option);
				if(limit >= 0) label->setScrollLimit(limit);
			}
		}
	}
}

// !LabelSetFontColor <label> <red> <green> <blue> OR !LabelSetFontColor <label> <hexcolor>
void BangSetFontColor(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16];
	char *buffers[] = {name, A, B, C};
	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);
	int color;

	if(numTokens >= 2)
	{
		#ifndef xLabelLight
		//----------------------------------------------------------------
			if (stricmp(name, "xdesktop") == 0)
			{
				for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				{
					if(numTokens >= 4)
					{
						int r = atoi(A);
						int g = atoi(B);
						int b = atoi(C);

						color = RGB(r, g, b);
					}
					else
					{
						color = strtol(A, NULL, 16);
						color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
					}

					(*i)->getFont()->setColor(color);
					(*i)->repaint();
				}
			}
			else
			{
		//----------------------------------------------------------------
		#endif
			Label *label = lookupLabel(name);

			if (label)
			{
				if(numTokens >= 4)
				{
					int r = atoi(A);
					int g = atoi(B);
					int b = atoi(C);

					color = RGB(r, g, b);

					string setcol = A;
					setcol.append( " " );
					setcol.append( B );
					setcol.append( " " );
					setcol.append( C );
					PrefixedLSSetVariable( name, "FontColor", setcol );
				}
				else
				{
					color = strtol(A, NULL, 16);
					color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

					PrefixedLSSetVariable( name, "FontColor", (string) A );
				}

				label->getFont()->setColor(color);
				label->repaint();
			}
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
					{
						if(numTokens >= 4)
						{
							int r = atoi(A);
							int g = atoi(B);
							int b = atoi(C);

							color = RGB(r, g, b);

							string setcol = A;
							setcol.append( " " );
							setcol.append( B );
							setcol.append( " " );
							setcol.append( C );
							PrefixedLSSetVariable( name, "FontColor", setcol );
						}
						else
						{
							color = strtol(A, NULL, 16);
							color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

							PrefixedLSSetVariable( name, "FontColor", (string) A );
						}

						(*i)->getFont()->setColor(color);
						(*i)->repaint();
					}
				}
			}
		#ifndef xLabelLight
		//----------------------------------------------------------------
			}
		//----------------------------------------------------------------
		#endif
	}
}

// !LabelSetFontShadowColor <label> <red> <green> <blue> OR !LabelSetFontColor <label> <hexcolor>
void BangSetFontShadowColor(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], A[16], B[16], C[16];
	char *buffers[] = {name, A, B, C};
	int numTokens = LCTokenize(pszArgs, buffers, 4, NULL);
	int color;

	if (numTokens == 1)
	{
		#ifndef xLabelLight
		//----------------------------------------------------------------
			if (stricmp(name, "xdesktop") == 0)
			{
				for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				{
					(*i)->getFont()->setShadow(false);
					(*i)->repaint();
					if ((*i)->getTransparencyMode() == 1)
						(*i)->updateRegion();
				}
			}
			else
			{
		//----------------------------------------------------------------
		#endif
			Label *label = lookupLabel(name);

			if (label)
			{
				label->getFont()->setShadow(false);
				label->repaint();
				if (label->getTransparencyMode() == 1)
					label->updateRegion();

				PrefixedLSSetVariable( name, "FontShadow", "false" );
			}
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
					{
						(*i)->getFont()->setShadow(false);
						(*i)->repaint();
						if ((*i)->getTransparencyMode() == 1)
							(*i)->updateRegion();

						PrefixedLSSetVariable( name, "FontShadow", "false" );
					}
				}
			}
		#ifndef xLabelLight
		//----------------------------------------------------------------
			}
		//----------------------------------------------------------------
		#endif
	}
	else if(numTokens >= 2)
	{
		#ifndef xLabelLight
		//----------------------------------------------------------------
			if (stricmp(name, "xdesktop") == 0)
			{
				for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				{
					if(numTokens >= 4)
					{
						int r = atoi(A);
						int g = atoi(B);
						int b = atoi(C);

						color = RGB(r, g, b);
					}
					else
					{
						color = strtol(A, NULL, 16);
						color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
					}

					(*i)->getFont()->setColor(color, true);
					(*i)->getFont()->setShadow(true);
					(*i)->repaint();
					if ((*i)->getTransparencyMode() == 1)
						(*i)->updateRegion();
				}
			}
			else
			{
		//----------------------------------------------------------------
		#endif
			Label *label = lookupLabel(name);

			if (label)
			{
				if(numTokens >= 4)
				{
					int r = atoi(A);
					int g = atoi(B);
					int b = atoi(C);

					color = RGB(r, g, b);

					string setcol = A;
					setcol.append( " " );
					setcol.append( B );
					setcol.append( " " );
					setcol.append( C );
					PrefixedLSSetVariable( name, "FontShadowColor", setcol );
				}
				else
				{
					color = strtol(A, NULL, 16);
					color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

					PrefixedLSSetVariable( name, "FontShadowColor", (string) A );
				}

				label->getFont()->setColor(color, true);
				label->getFont()->setShadow(true);
				label->repaint();
				if (label->getTransparencyMode() == 1)
					label->updateRegion();

				PrefixedLSSetVariable( name, "FontShadow", "true" );
			}
			else
			{
				LabelListIterator i;
				for( i = labelList.begin(); i != labelList.end(); i++)
				{
					if( stricmp((*i)->labelGroup.c_str(), name) == 0 )
					{
						if(numTokens >= 4)
						{
							int r = atoi(A);
							int g = atoi(B);
							int b = atoi(C);

							color = RGB(r, g, b);

							string setcol = A;
							setcol.append( " " );
							setcol.append( B );
							setcol.append( " " );
							setcol.append( C );
							PrefixedLSSetVariable( name, "FontShadowColor", setcol );
						}
						else
						{
							color = strtol(A, NULL, 16);
							color = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

							PrefixedLSSetVariable( name, "FontShadowColor", (string) A );
						}

						(*i)->getFont()->setColor(color, true);
						(*i)->getFont()->setShadow(true);
						(*i)->repaint();
						if ((*i)->getTransparencyMode() == 1)
							(*i)->updateRegion();

						PrefixedLSSetVariable( name, "FontShadow", "true" );
					}
				}
			}
		#ifndef xLabelLight
		//----------------------------------------------------------------
			}
		//----------------------------------------------------------------
		#endif
	}
}

// !LabelSetText <label> <text>
void BangSetText(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], text[MAX_LINE_LENGTH];
	char *buffers[] = {name, text};
	int numTokens = LCTokenize(pszArgs, buffers, 2, NULL);

	if(numTokens >= 1)
	{
		Label *label = lookupLabel(name);

		if(label)
			label->setText(((numTokens >= 2) ? text : ""), true);
	}
}

// !LabelToggle <label>
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	#ifndef xLabelLight
	//----------------------------------------------------------------
		if (stricmp(pszArgs, "xdesktop") == 0)
		{
			for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
			{
				if((*i)->isVisible())
					(*i)->hide();
				else
					(*i)->show();
			}
		}
		else
		{
	//----------------------------------------------------------------
	#endif
		Label *label = lookupLabel(pszArgs);
		if(label)
		{
			if(label->isVisible())
				label->hide(true);
			else
				label->show(true);
		}
		else
		{
			LabelListIterator lit;
			for( lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), pszArgs) == 0 )
				{
					if((*lit)->isVisible())
						(*lit)->hide(true);
					else
						(*lit)->show(true);
				}
			}

			#ifndef xLabelLight
			//----------------------------------------------------------------
				char group[MAX_LINE_LENGTH];
				LPCSTR nextGroup = pszArgs;

				while (GetToken(nextGroup, group, &nextGroup, false))
				{
					for( lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
					{
						if( stricmp((*lit)->labelGroup.c_str(), group) == 0 || stricmp((*lit)->labelGroup.c_str(), "*") == 0 )
						{
							if((*lit)->isVisible())
								(*lit)->hide();
							else
								(*lit)->show();
						}
					}
				}
			//----------------------------------------------------------------
			#endif
		}
	#ifndef xLabelLight
	//----------------------------------------------------------------
		}
	//----------------------------------------------------------------
	#endif
}

// !LabelUpdate <label>
void BangUpdate(HWND hwndCaller, LPCSTR pszArgs)
{
	Label *label = lookupLabel(pszArgs);
	if(label)
		label->update();
	else
	{
		LabelListIterator i;
		for( i = labelList.begin(); i != labelList.end(); i++)
		{
			if( stricmp((*i)->labelGroup.c_str(), pszArgs) == 0 )
				(*i)->update();
		}
	}
}

// !LabelSetAlpha <label> <value>
void BangSetAlpha(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], factor[16];
	char *buffers[] = {name, factor};

	bool wasVisible = false;

	if(LCTokenize(pszArgs, buffers, 2, NULL) == 2)
	{
		#ifndef xLabelLight
		//----------------------------------------------------------------
			if (stricmp(name, "xdesktop") == 0)
			{
				
				for(LabelListIterator i = desktoplabelList.begin(); i != desktoplabelList.end(); i++)
				{
					if ( !(*i)->isLayeredFlagSet() )
					{
						if ((*i)->isVisible())
						{
							wasVisible = true;
							(*i)->hide();
						}
						else 
							wasVisible = false;
					}

					(*i)->setAlpha(atoi( factor ));

					if (wasVisible)
						(*i)->show();
				}
			}
			else
			{
		//----------------------------------------------------------------
		#endif
			Label *label = lookupLabel(name);
			if(label)
			{
				if ( !label->isLayeredFlagSet() )
				{
					if (label->isVisible())
					{
						wasVisible = true;
						label->hide();
					}
					else 
						wasVisible = false;
				}

				label->setAlpha(atoi( factor ));

				if (wasVisible)
					label->show();
			}
			else
			{
				LabelListIterator lit;
				for( lit = labelList.begin(); lit != labelList.end(); lit++)
				{
					if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
					{
						if ( !(*lit)->isLayeredFlagSet() )
						{
							if ((*lit)->isVisible())
							{
								wasVisible = true;
								(*lit)->hide();
							}
							else 
								wasVisible = false;
						}

						(*lit)->setAlpha(atoi( factor ));

						if (wasVisible)
							(*lit)->show();
					}
				}
			}
		#ifndef xLabelLight
		//----------------------------------------------------------------
			}
		//----------------------------------------------------------------
		#endif
	}
}

// !LabelGhosted <label> <on|true, off|false, toggle>
void BangGhosted(HWND hwndCaller, LPCSTR pszArgs)
{
	char name[64], option[32];
	char *buffers[] = {name, option};

	LCTokenize(pszArgs, buffers, 2, NULL);

	Label *label = lookupLabel(name);
	if(label)
	{
		if (stricmp(option, "true") == NULL || stricmp(option, "on") == NULL)
			label->setGhosted(true);
		else if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
			label->setGhosted(false);
		else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
			label->setGhosted(!label->isGhosted());
	}
	else
	{
		for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
		{
			if( stricmp((*lit)->labelGroup.c_str(), name) == 0 )
			{
				if (stricmp(option, "true") == NULL || stricmp(option, "on") == NULL)
					(*lit)->setGhosted(true);
				else if (stricmp(option, "false") == NULL || stricmp(option, "off") == NULL)
					(*lit)->setGhosted(false);
				else if (stricmp(option, "toggle") == NULL || stricmp(option, "switch") == NULL)
					(*lit)->setGhosted(!(*lit)->isGhosted());
			}
		}
	}
}

#ifndef xLabelLight
//----------------------------------------------------------------
//----------------------------------------------------------------

	void BangSCOnBottom(HWND hwndCaller, LPCSTR pszArgs)
	{
		char group[MAX_LINE_LENGTH];
		LPCSTR nextGroup = pszArgs;

		while (GetToken(nextGroup, group, &nextGroup, false))
		{
			for(LabelListIterator lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), group) == 0 || stricmp((*lit)->labelGroup.c_str(), "*") == 0 )
					(*lit)->setAlwaysOnTop(false);
			}
		}
	}


	void BangSCToggleOnTop(HWND hwndCaller, LPCSTR pszArgs)
	{
		char group[MAX_LINE_LENGTH];
		LPCSTR nextGroup = pszArgs;

		while (GetToken(nextGroup, group, &nextGroup, false))
		{
			for(LabelListIterator lit = shortcutlabelList.begin(); lit != shortcutlabelList.end(); lit++)
			{
				if( stricmp((*lit)->labelGroup.c_str(), group) == 0 || stricmp((*lit)->labelGroup.c_str(), "*") == 0 )
					(*lit)->setAlwaysOnTop(!(*lit)->isAlwaysOnTop());
			}
		}
	}

	//TEXTEDIT 2.52 SECTION:
	//=========================================================
	// getToken grabs a token and clobbers info that is parsed
	//=========================================================
	void getToken(char line[MAX_LINE_LENGTH], char temp[MAX_LINE_LENGTH], bool escape_codes /*= true*/)
	{
		int offset = 0, i;
		bool endquote = false;
		char ENCAP = '@';

		while (line[offset] == ' ')
		{
			offset++;
		}

		if (line[offset] == ENCAP)
		{
			offset++;
			endquote = true;
		}

		for (i = 0; i+offset < int(strlen(line)); ++i)
		{
			// Copy else if block to add more escape codes.
			if (escape_codes && (line[i+offset] == '\\' && int(strlen(line)) != i+offset+1))
			{
				if (line[i+offset+1] == ENCAP || line[i+offset+1] == '\\')
				{
					offset++;
					temp[i] = line[i+offset];
					continue;
				}
				// Bang escape code below. Present in 2.0 from Tony Chang
				else if (line[i+offset+1] == '^')
				{
					offset++;
					temp[i] = '!';
					continue;
				}
				// EVar escape code below. Present in 2.0 from Tony Chang
				else if (line[i+offset+1] == '#')
				{
					offset++;
					temp[i] = '$';
					continue;
				}
				// Comment escape code below. Added in 2.1l by leaf r&d
				else if (line[i+offset+1] == '~')
				{
					offset++;
					temp[i] = ';';
					continue;
				}
				// Quotation mark escape code below. Added in 2.2l by leaf r&d
				else if (line[i+offset+1] == '=')
				{
					offset++;
					temp[i] = '"';
					continue;
				}
				// Clipboard escape code
				/*else if (line[i+offset+1] == 'c')
				{
					HGLOBAL hMem;
					LPVOID pvMem;

					if(IsClipboardFormatAvailable(CF_TEXT))
					{
						if(OpenClipboard(NULL))
						{
							hMem = GetClipboardData(CF_TEXT);

							if(hMem)
							{
								char pszBuffer[512];
								pvMem = GlobalLock(hMem);
								lstrcpyn(pszBuffer, (LPCTSTR) pvMem, 512);
								temp[i] = '\0';
								offset += min(strlen(pszBuffer), 512-strlen(temp));
								strncat(temp, pszBuffer, 512-strlen(temp));
								i = strlen(temp)+1;
								GlobalUnlock(hMem);
							}

							CloseClipboard();
						}
					}
				}*/
			}

 			if (line[i+offset] == ENCAP && endquote)
	  			break;
			else if (line[i+offset] == ' ' && !endquote)
				break;
			else
				temp[i] = line[i+offset];
		}
		temp[i] = '\0';

		if (i+offset+1 > int(strlen(line)))
		{
			line[0] = '\0';
		}
		else
		{
			strcpy(line, line+i+offset+1);
		}
	}

	//=========================================================
	// Bang command handling
	//=========================================================

	void BangTextAppend(HWND hwndCaller, LPCSTR pszArgs)
	{
		char filename[MAX_LINE_LENGTH], str[MAX_LINE_LENGTH];
		char rest[MAX_LINE_LENGTH];
		strcpy(rest, pszArgs);

		getToken(rest, filename, false);
		getToken(rest, str, false);

		ofstream file;
		file.open(filename, /* ios::nocreate | */ ios::app);

		if (!file.fail())
		{
			file << str << endl;
		}

		file.close();
	}

	void BangTextDelete(HWND hwndCaller, LPCSTR pszArgs)
	{
		char filename[MAX_LINE_LENGTH], regex[MAX_LINE_LENGTH], input[MAX_LINE_LENGTH];
		char rest[MAX_LINE_LENGTH];
		regexp *r;
		strcpy(rest, pszArgs);

		getToken(rest, filename, false);
		getToken(rest, regex, true);

		r = regcomp(regex);

		ifstream file;
		file.open(filename/*, ios::nocreate*/);        //removed because dev-cpp wont compile it... feel free to recompile to get it to work

		if (!file.fail())
		{
			list<string> l;
			string s;
			while (file.getline(input, MAX_LINE_LENGTH))
			{
				if (!regexec(r, input))
				{
					s = input;
					l.push_back(s);
				}
				else
				{
					if ( userConfirmation || userDebug )
					{
						string message = "Do you really want to delete\n\n";
						message.append( input );
						message.append ( "\n\n in the file\n\n" );
						message.append ( filename );
						if ( userDebug )
						{
							message.append( "\n\nSearch String was:\n\n" );
							message.append( regex );
						}
						if ( MessageBox(NULL, message.c_str(), "User Confirmation", MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDNO )
						{
							s = input;
							l.push_back(s);
						}
					}
				}
			}
			file.close();
			ofstream outfile;
			outfile.open(filename);
			for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
			{
				outfile << lci->data() << endl;
			}
			l.clear();
		}
	}

	void BangTextReplace(HWND hwndCaller, LPCSTR pszArgs)
	{
		char filename[MAX_LINE_LENGTH], regex[MAX_LINE_LENGTH], input[MAX_LINE_LENGTH], replace[MAX_LINE_LENGTH];
		char rest[MAX_LINE_LENGTH], buf[MAX_LINE_LENGTH];
		regexp *r;
		strcpy(rest, pszArgs);

		getToken(rest, filename, false);
		getToken(rest, regex, true);
		getToken(rest, replace, true);

		r = regcomp(regex);

		ifstream file;
		file.open(filename/*, ios::nocreate*/);        //removed because dev-cpp wont compile it... feel free to recompile to get it to work

		bool found = false;
		if (!file.fail())
		{
			
			list<string> l;
			string s;
			while (file.getline(input, MAX_LINE_LENGTH))
			{
				if ( !found && regexec(r, input) )
				{
					regsub(r, replace, buf);

					if ( !userConfirmation && !userDebug)
					{
						s = buf;
						l.push_back(buf);
						found = true;
					}
					else
					{
						string message = "Do you want to replace\n\n";
						message.append( input );
						message.append( "\nwith\n" );
						message.append ( buf );
						message.append ( "\n\n in the file\n\n" );
						message.append ( filename );
						if ( userDebug )
						{
							message.append( "\n\nSearch String was:\n" );
							message.append( regex );
							message.append( "\n\nReplace String was:\n" );
							message.append( replace );
						}
						if ( MessageBox(NULL, message.c_str(), "User Confirmation", MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION ) == IDYES )
						{
							found = true;
							s = buf;
							l.push_back(buf);
						}
						else
						{
							found = true;
							s = input;
							l.push_back(input);
						}
					}
				}
				else
				{
					s = input;
					l.push_back(input);
				}
			}
			file.close();
			ofstream outfile;
			outfile.open(filename);
			for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
			{
				outfile << lci->data() << endl;
			}
			l.clear();
		}

		if ( !found )
		{
			string message = "The given Search String\n\n";
			message.append( regex );
			message.append ( "\n\nwasn't found in the file\n\n" );
			message.append ( filename );
			message.append ( "\n\nReplace canceled!" );
			MessageBox(NULL, message.c_str(), "User Confirmation", MB_OK | MB_SETFOREGROUND | MB_ICONEXCLAMATION);
		}
	}

	void BangTextReplaceAll(HWND hwndCaller, LPCSTR pszArgs)
	{
		char filename[MAX_LINE_LENGTH], regex[MAX_LINE_LENGTH], input[MAX_LINE_LENGTH], replace[MAX_LINE_LENGTH];
		char rest[MAX_LINE_LENGTH], buf[MAX_LINE_LENGTH];
		regexp *r;
		strcpy(rest, pszArgs);

		getToken(rest, filename, false);
		getToken(rest, regex, true);
		getToken(rest, replace, true);

		r = regcomp(regex);

		ifstream file;
		file.open(filename/*, ios::nocreate*/);        //removed because dev-cpp wont compile it... feel free to recompile to get it to work

		if (!file.fail())
		{
			list<string> l;
			string s;
			while (file.getline(input, MAX_LINE_LENGTH))
			{
				if (regexec(r, input))
				{
					s = buf;
					l.push_back(buf);
				}
				else
				{
					s = input;
					l.push_back(input);
				}
			}
			file.close();
			ofstream outfile;
			outfile.open(filename);
			for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
			{
				outfile << lci->data() << endl;
			}
			l.clear();
		}
	}

	void BangTextSaveEvar(HWND hwndCaller, LPCSTR pszArgs)
	{
		char filename[MAX_LINE_LENGTH] = {0};
		char str[MAX_LINE_LENGTH] = {0};
		char val[MAX_LINE_LENGTH] = {0};
		char rest[MAX_LINE_LENGTH] = {0};

		strcpy(rest, pszArgs);

		getToken(rest, filename, false);
		getToken(rest, str, false);
		getToken(rest, val, false);

		string file = filename;
		string EvarToSave = str;
		string result = "";
			
		int length = EvarToSave.length();
		int i = 0;
			
		while(i < length)
		{
			result.append(1, '[');
			result.append(1, tolower(EvarToSave[i]));
			result.append(1, toupper(EvarToSave[i++]));
			result.append(1, ']');
		}
			
		EvarToSave = result;

		result = val;
		if ( result.empty() )
		{
			result = "$";
			result.append( str );
			result.append( "$" );
			VarExpansion( str, result.c_str() );
		}
		else
			strcpy( str, val);

		result = str;

		BangTextReplace(NULL, ("@" + file + "@ @[		]*(" + EvarToSave + ").*@ @\\1 \"" + result + "\"@").c_str());
	}

//----------------------------------------------------------------
//----------------------------------------------------------------
#endif

//Parse Current Evars In Bang
//!ParseEvars !anybang with evars escape code '%#' and TextEscape Sequences with %[...]%
void BangParseEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	CString helper;
	CString helper1;
	string result = pszArgs;
	while ( strstr(result.c_str(), "%#") != NULL )
	{
		result.replace(result.find("%#"), 2, "$", 1);
	}

	while ( strstr(result.c_str(), "%|#") != NULL )
	{
		result.replace(result.find("%|#"), 3, "$", 1);
	}

	helper1 = result.c_str();

	while ( strstr(helper1, "%[") != NULL && strstr(helper1, "]%") != NULL)
	{
		helper = helper1;
		helper = helper.Right(helper.GetLength()-helper.Find("%[")-1);
		helper = helper.Left(helper.Find("]%")+1);

		bool dynamic = false;
		if ( stricmp(helper, "[labelname]") == NULL)
		{
			for(LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if( (*lit)->gethWnd() == hwndCaller)
				{
					helper1.Replace( ("%"+helper+"%"), (systemInfo->processLabelText( (string)helper, (*lit), &dynamic)).c_str() );
					break;
				}
			}
		}
		else
			helper1.Replace( ("%"+helper+"%"), (systemInfo->processLabelText( (string)helper, NULL, &dynamic)).c_str() );
	}

	xLSExecute(hwndCaller, helper1, SW_SHOWNORMAL);
}

//Set Evar
//!SetEvar "'$'evar'$'" "value"
void BangSetEvars(HWND hwndCaller, LPCSTR pszArgs)
{
	char evar[64], value[MAX_PATH_LENGTH];
	char *buffers[] = {evar};

	if(LCTokenize(pszArgs, buffers, 1, value) >= 1)
	{
		PrefixedLSSetVariable( (string)evar, "", (string)value );
	}
}

//Play Wav
//!PaySound "Path to wav|mp3|mid"
void BangPlaySound(HWND hwndCaller, LPCSTR pszArgs)
{
	char sound[MAX_PATH_LENGTH];
	char *buffers[] = {sound};

	LCTokenize(pszArgs, buffers, 1, NULL);

	CString m_Path = sound;
	m_Path.MakeLower();

	if (m_Path.Find(".wav") || m_Path.Find(".mp3") || m_Path.Find(".mid"))
	{
		if (m_hWndSoundFile != NULL)
		{
			MCIWndStop(m_hWndSoundFile);
			MCIWndDestroy(m_hWndSoundFile);
			m_hWndSoundFile = NULL;
		}

		HANDLE hFindHandle;
		WIN32_FIND_DATA FindData;

		hFindHandle = FindFirstFile(m_Path, &FindData);

		if(INVALID_HANDLE_VALUE != hFindHandle)
		{
			m_hWndSoundFile = MCIWndCreate(NULL, NULL, WS_MINIMIZE | MCIWNDF_NOMENU, m_Path);
			
			MCIWndPlay(m_hWndSoundFile);

			SetTimer(messageHandler, NULL, 1000, 0);
		}

		FindClose(hFindHandle);
	}
	else if (m_hWndSoundFile != NULL)
	{
		MCIWndStop(m_hWndSoundFile);
		MCIWndDestroy(m_hWndSoundFile);
		m_hWndSoundFile = NULL;
	}
}